 
/*
 *      (c) Copyright 1991 Sun Microsystems, Inc. Sun design patents
 *      pending in the U.S. and foreign countries. See LEGAL NOTICE
 *      file for terms of the license.
 */

/*
 * Name:	fifo.c
 *
 * Synopsis:	void 	fifo_init(fifo_t * fifo, int type, void * arg);
 *
 * int 	fifo_enter(fifo_t * fifo, void (*func)(), void * arg);
 *
 * int  	fifo_tryenter(fifo_t * fifo);
 *
 * void 	fifo_exit(fifo_t * fifo);
 *
 * void	fifo_yield(fifo_t * fifo, unsigned int count)
 *
 * int 	fifo_num_waiters(fifo_t * fifo);
 *
 *
 * Description:	These routines provide a fifo recursive mutex facility for
 * threads.  Provision is made to support calling a user supplied signaling
 * routine before threads block on the fifo and urgent acquisition of the
 * token.
 *
 * Fifo_init initializes a fifo_t data structure.  Note that fifos may be
 * statically initialized as zero as well.
 *
 * A thread calling fifo_enter will block until it can have exclusive access to
 * the fifo.  If the calling thread already has the fifo, an internal
 * recursion count is incremented and the routine returns immediately.  If
 * the thread is about to block and the user function pointer func is
 * non-NULL, the specified routine is called with the user-specified argument
 * arg. Note that the call is performed while still holding the fifo's
 * internal lock, so any race conditions are avoided. Threads blocked on the
 * fifo will obtain the fifo in order.
 *
 * Fifo_tryenter is used by threads which do not want to block on waiting for
 * the fifo.  If the fifo_tryenter returns 1, the calling thread now has the
 * fifo (or already did).  If it returns 0, the fifo is held by some other
 * thread.
 *
 * Fifo_exit is called by threads releasing the fifo.  The recursion count is
 * decremented and if zero the next thread waiting (if any) is alerted that
 * it now has the fifo.
 *
 * Fifo_yield is used to allow a specified number of threads to acquire the
 * token.  It allows the holder of the fifo to release the fifo (regardless
 * of the level of recursion), and re-insert itself into the queue of waiting
 * threads.  The position in the queue is specified by count, 0 being after
 * the 0th waiting thread and so forth.  When the calling thread resume
 * execution, it once again holds the fifo at the same level of recursion
 * which which it entered fifo_yield.
 *
 * Fifo_num_waiters returns the number of threads waiting for the fifo. */

#include <assert.h>
#include <thread.h>
#include "fifo.h"

/*
 * this routine is used to initialize a fifo before use note that fifos may
 * be staticly allocated successfully
 */

void
fifo_init(fifo_t * fifo, int type, void *arg)
{
	fifo->head = fifo->tail = NULL;	/* indicate no one waiting */
	mutex_init(&fifo->mutex, 0, 0);	/* prep mutex */
	fifo->holder = 0;	/* no holder */
	fifo->waiting = 0;	/* no one waiting either */
}

/*
 * This routine attempts to acquire the fifo.  If not already held, the fifo
 * is acquired immediately; otherwise the calling thread goes to sleep after
 * calling the user-supplied func with arg as the single argument.  The
 * calling thread is placed at the end of the queue of waiting threads.
 * Fifo_signal_enter returns 1 if the user-supplied function was called, else
 * 0.
 */

int fifo_enter(fifo_t * fifo, void (*func) (), void *arg)
{
	int ret = 0;
	q_entry_t my_entry;	/* allocate q entry on stack */

	mutex_enter(&fifo->mutex);

	my_entry.next = NULL;
	my_entry.thread_id = thread_get_id();
	if (fifo->holder) {	/* uh-oh - someone already holds the fifo */

		if (my_entry.thread_id == fifo->holder) {	/* its me! */
			fifo->recursion++;
			goto exit;
		}
		cv_init(&my_entry.cv, 0, 0);
		my_entry.ok_to_run = 0;

		if (fifo->head) {
			fifo->tail->next = &my_entry;
			fifo->tail = &my_entry;
		} else {	/* I'm the first waiter... */
			fifo->head = &my_entry;
			fifo->tail = &my_entry;
		}

		fifo->waiting++;

		if (func) {
			(*func) (arg);
			ret++;
		}
		while (!my_entry.ok_to_run)
			cv_wait(&my_entry.cv, &fifo->mutex);
	}
	fifo->holder = my_entry.thread_id;	/* Its mine! */

exit:
	mutex_exit(&fifo->mutex);
	return (ret);
}

/*
 * Yield to next count threads on queue.
 */

void
fifo_yield(fifo_t * fifo, unsigned int count)
{
	q_entry_t my_entry;
	int save_recursion;

	mutex_enter(&fifo->mutex);

	assert(fifo->holder == thread_get_id());

	/*
	 * if fifo->head is NULL,
	 * nothing is waiting, just
	 * return
	 */

	if (fifo->head) {
		/*
		 * save holder's fifo
		 * context
		 */

		my_entry.thread_id = fifo->holder;
		save_recursion = fifo->recursion;

		/*
		 * set up to run other
		 * thread
		 */
		fifo->holder = fifo->head->thread_id;
		fifo->recursion = 0;
		fifo->head->ok_to_run = 1;
		cv_signal(&(fifo->head->cv));
		fifo->head = fifo->head->next;

		/*
		 * set up to block
		 * calling thread
		 */
		cv_init(&my_entry.cv, 0, 0);
		my_entry.ok_to_run = 0;
		my_entry.next = NULL;

		if (!fifo->head) {	/* no other threads - just add me */
			fifo->head = &my_entry;
			fifo->tail = &my_entry;
		} else {
			q_entry_t *insert_after = fifo->head;

			while (count-- && insert_after->next) {
				insert_after = insert_after->next;
			}

			if (!(my_entry.next = insert_after->next))
				fifo->tail = &my_entry;

			insert_after->next = &my_entry;
		}

		/*
		 * block until it is
		 * original thread's
		 * turn to run again
		 */
		while (!my_entry.ok_to_run)
			cv_wait(&my_entry.cv, &fifo->mutex);
		/*
		 * restore original
		 * thread's fifo
		 * context
		 */
		fifo->recursion = save_recursion;
		fifo->holder = my_entry.thread_id;
	}
	mutex_exit(&fifo->mutex);
}

/*
 * fifo_try_enter attempts to acquire the fifo without blocking.  If
 * succesful, it returns 1.  Else it returns 0.
 */

int
fifo_tryenter(fifo_t * fifo)
{
	int ret = 0;

	mutex_enter(&fifo->mutex);

	if (!fifo->holder) {	/* got it ! */
		fifo->holder = thread_get_id();
		ret = 1;
	} else if (fifo->holder == thread_get_id()) {
		fifo->recursion++;	/* already got it! */
		ret = 1;
	}
	mutex_exit(&fifo->mutex);
	return (ret);
}

/*
 * Fifo_exit releases the fifo
 */

void
fifo_exit(fifo_t * fifo)
{
	mutex_enter(&fifo->mutex);

	if (fifo->recursion > 0) {
		fifo->recursion--;
	} else if (fifo->head) {
		fifo->holder = fifo->head->thread_id;
		fifo->waiting--;
		fifo->head->ok_to_run = 1;	/* assure him this is genuine */
		cv_signal(&(fifo->head->cv));	/* let him know to wake up  */

		if ((fifo->head = fifo->head->next) == NULL) {
			/*
			 * clean up the
			 * queue
			 */
			fifo->tail = NULL;
		}
	} else {
		fifo->holder = NULL;	/* no waiters */
	}

	mutex_exit(&fifo->mutex);
}

/*
 * this routine returns the number of threads waiting to acquire the fifo
 */

int
fifo_num_waiters(fifo_t * fifo)
{
	int ret;

	mutex_enter(&fifo->mutex);
	ret = fifo->waiting;
	mutex_exit(&fifo->mutex);
	return (ret);
}

#ifdef TESTFIFO


#define	LOOPS 50

#include <stropts.h>
#include <poll.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>


static fifo_t m;
static fifo_t *mp = &m;
static int fildes[2];
static struct sigaction action;
static thread_id_t main_thread_id;

main(int argc, char *argv[])
{
	void main_thread();
	int count;

	if (argc == 1)
		count = 1;
	else
		count = atoi(argv[1]);

	thread_create(NULL, 0, main_thread, (caddr_t) count, THREAD_NEW_LWP);

}


void
main_thread(int count)
{
	int i;

	void child();

	if (pipe(fildes))
		perror("pipe");

	main_thread_id = thread_get_id();

	fifo_init(mp, 0, 0);

	fifo_enter(mp, NULL, NULL);

	for (i = 0; i < count; i++)
		thread_create(NULL, 0, child, NULL, THREAD_BIND_LWP);

	for (;;) {
		struct pollfd pollfd;
		int ret;

		pollfd.fd = fildes[0];
		pollfd.events = POLLIN;
		pollfd.revents = 0;

		ret = poll(&pollfd, 1, 3000);

		switch (ret) {
		case 1:
			fifo_yield(mp, 5);

			break;

		case 0:
			printf("poll: timeout\n");
			return;
			break;

		case -1:
			if (errno == EINTR)
				printf("poll: interrupted system call\n");
			break;
		default:
			printf("Poll: returned %d???\n", ret);
			break;
		}
	}
}

void
pinger()
{
	int i = thread_get_id();

	write(fildes[1], &i, sizeof (i));
}

void
child()
{
	int loops;

	for (loops = 0; loops < LOOPS; loops++) {
		int i, j, pid;

		if (fifo_enter(mp, pinger, 0))
			if ((j = readable(fildes[0])) >= 4) {
				read(fildes[0], &i, 4);
				assert(i == thread_get_id());
			} else
				printf("readable returned %d\n", j);
		fifo_exit(mp);
	}
}

static int
readable(fildes)
	int fildes;
{
	struct stat buf;

	if (fstat(fildes, &buf) == -1)
		perror("readable");

	return (buf.st_size);
}


#endif				/* TESTFIFO */
